/*
 * AnnotateImage.java
 *
 * Release: 1.0
 * Build:   20021023
 *
 * Copyright 2002 Gary Gregory and Deborah Lewis. All Rights Reserved.
 *
 * This software is the proprietary information of Glaivestone Guild,
 * founded 2002 by Gary Gregory and Deborah Lewis.
 * Use is subject to license terms.
 */

package com.glaivestone.ant.taskdefs;

import org.apache.tools.ant.Task; //superclass

// Java base class library
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

// Java UI library
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;

// in JDK and JRE, but not part of core java API's
import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGImageDecoder;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

// Ant
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;

/**
 * Ant task to annotate an image with text.
 * 
 * <p>Refer to the task documentation for information about attributes and usage of this task.
 * 
 * @author Gary Gregory
 * @author Deb Lewis
 */
public class AnnotateImage extends Task {

    //----------------------------------------------------------------------------
    // CONSTANTS

	public static final String IMAGE_TYPE_JPEG = "jpeg";
	//TO DO: public static final String IMAGE_TYPE_GIF = "gif";

	public static final String DEFAULT_IMAGE_TYPE = IMAGE_TYPE_JPEG;
	public static final String[] SUPPORTED_IMAGE_TYPES = {
		IMAGE_TYPE_JPEG,
		//IMAGE_TYPE_GIF,  // need a writer...
	};

	public static final int DEFAULT_FONT_SIZE = -1;

	public static final Map colorNameTable = constructColorNameTable();

    //----------------------------------------------------------------------------
    // INSTANCE VARIABLES

	protected String imageType;
	protected File imageFile;
	protected File imageSource;

	protected boolean useDefaultFont;
	protected String fontName;
	protected int fontStyle;
	protected int fontSize;

	protected Color fontColor;

	protected List annotations;

    //----------------------------------------------------------------------------
    // CLASS INITIALIZATION

	protected static Map constructColorNameTable() {
		Map aTable = new HashMap();
		aTable.put( "black", Color.black );
		aTable.put( "white", Color.white );
		aTable.put( "red", Color.red );
		aTable.put( "blue", Color.blue );
		aTable.put( "yellow", Color.yellow );
		aTable.put( "green", Color.green );
		return aTable;
	}

    //----------------------------------------------------------------------------
    // TASK INITIALIZATION

    /**
     * Called by the project to let the task initialize properly. 
	 *
	 * @throws BuildException if someting goes wrong with the task initialization
     */
	public void init() 
		throws BuildException
    {
		super.init();
		this.setType( DEFAULT_IMAGE_TYPE );
		this.useDefaultFont = true;
		this.fontName = null;
		this.fontStyle = Font.PLAIN;
		this.fontSize = DEFAULT_FONT_SIZE;
		this.fontColor = Color.black;
		this.annotations = new ArrayList();
    }

    //----------------------------------------------------------------------------
    // TASK ATTRIBUTE SETTERS

	public void setColor( String colorSpec ) {
		//TBD: allow "rgb(rInt, gInt, bInt)" encoding...
		Color aColor = (Color) colorNameTable.get( colorSpec );
		if ( aColor == null ) {
			throw new BuildException(
				"Unknown annotation color '" + colorSpec + "'",
				this.getLocation() );
		}
		this.fontColor = aColor;
	}

	/*
	 * Image encoding type.
	 *
	 * @throws BuildException if not a supported image type
	 */
	public void setType( String anImageType ) {
		this.imageType = this.normalizeImageType( anImageType );;
	}

	/**
	 * Image file to annotate.  This is the file created by the task;
	 * may also be the source of the image for an in-place annotation update.
	 */
	public void setImagefile( File aFile ) {
		this.imageFile = aFile;
	}

	/**
	 * Source of the image to annotate.
	 *
	 * @throws BuildException if the specified source image file does not exist
	 */
	public void setImagesource( File aFile ) {
		if ( ! aFile.exists() ) {
			throw new BuildException( 
				"Image source file does not exist: " + aFile,
				this.getLocation() );
		}
		this.imageSource = aFile;
	}

	/**
	 * Specify that the name of the font to be used.
	 */
	public void setFont( String fontName ) {
		if ( fontName == null || ( fontName = fontName.trim() ).length() == 0 ) {
			//should this be reported as a build spec error???
			return;
		}
		this.useDefaultFont = false;
		this.fontName = fontName;
	}

	/**
	 * Specify that bold font style should be used.
	 */
	public void setBold( boolean useBold ) {
		if ( useBold ) {
			this.useDefaultFont = false;
			this.fontStyle = this.fontStyle + Font.BOLD;
		}
	}

	/**
	 * Specify that italic font style should be used.
	 */
	public void setItalic( boolean useItalic ) {
		if ( useItalic ) {
			this.useDefaultFont = false;
			this.fontStyle = this.fontStyle + Font.ITALIC;
		}
	}

	/**
	 * Set the font size in points to use.
	 */
	public void setFontsize( int fontSize ) {
		this.fontSize = fontSize;
	}

    //----------------------------------------------------------------------------
    // NESTED ELEMENT SETTERS

	public void addConfiguredAnnotation( ImageAnnotation anAnnotation ) {
		if ( anAnnotation.getLeft() < 0 || anAnnotation.getTop() < 0 ) {
			throw new BuildException(
				"Text annotation requires upper left corner {x, y} position",
				this.getLocation() );
		}
		if ( ! anAnnotation.hasText() ) {
			throw new BuildException(
				"Text annotation requires text content",
				this.getLocation() );
		}
		this.annotations.add( anAnnotation );
	}

    //----------------------------------------------------------------------------
    // TASK EXECUTION

	/**
	 * Called by the project to let the task do it's work. This method may be 
	 * called more than once, if the task is invoked more than once. For example, 
	 * if target1 and target2 both depend on target3, then running 
	 * "ant target1 target2" will run all tasks in target3 twice.
	 *
	 * @throws BuildException if someting goes wrong with the build
	 */
	public void execute() 
		throws BuildException 
	{

		BufferedImage anImage; 

		this.validateTaskSpecification();

		this.log( 
			"Annotating image " + this.imageFile.getAbsolutePath() );
		if ( this.imageSource != null ) {
			this.log( "...from source " + this.imageSource.getAbsolutePath() );
		}

		try {
			anImage = this.loadImage();
		} catch( IOException ex ) {
			throw new BuildException( "Unable to load image: " + ex.getMessage(), ex );
		}

		this.drawAnnotations( anImage );

		try {
			this.saveImage( anImage );
		} catch( IOException ex ) {
			throw new BuildException( "Unable to save image: " + ex.getMessage(), ex );
		}

	} //execute

	/**
	 * Draw the text annotations onto the image.
	 */
	protected void drawAnnotations( BufferedImage anImage ) {

		Graphics2D gc;

		gc = anImage.createGraphics();
		try {
			this.initFont( gc );
			gc.setColor( this.fontColor );
			this.drawAnnotations( gc );
		} finally {
			gc.dispose();
		}

	} //drawAnnotations

	/**
	 * Draw the text annotations onto the image using the graphics context.
	 * 
	 * <p>Font height is the distance between successive baselines:<br>
	 * <pre>   descent + leading + ascent</pre><br>
	 * <i>(Core Java Vol. 1 - Horstmann and Cornell)</i>
	 * 
	 * <p>Note that we use max ascent/descent here to ensure we compute
	 * the maximum possible screen area (perhaps excessive, but safe).
	 */
	protected void drawAnnotations( Graphics2D gc ) {

		FontMetrics fm = gc.getFontMetrics();

		int heightAboveBaseline = fm.getLeading() + fm.getMaxAscent();
		int lineHeight = heightAboveBaseline + fm.getMaxDescent();

		// text drawing is done from the (x,y) position of the baseline
		// of the leftmost character in the line
		int nAnnotations = 0;
		for (Iterator iter = this.annotations.iterator(); iter.hasNext();) {
			ImageAnnotation anAnnotation = (ImageAnnotation) iter.next();
			int xBaseline = anAnnotation.getLeft();
			int yBaseline = anAnnotation.getTop() + heightAboveBaseline;
			for ( Iterator textIter = anAnnotation.getLines().iterator(); textIter.hasNext(); ) {
				String line = (String) textIter.next();
				this.log( "...(" + xBaseline + "," + yBaseline + ") '" + line + "'", Project.MSG_VERBOSE );
				gc.drawString( line, xBaseline, yBaseline );
				yBaseline += lineHeight;
				nAnnotations++;
			}
		}
		this.log( "..." + nAnnotations + " annotations inserted", Project.MSG_VERBOSE );

	} //drawAnnotations

	/**
	 * Set the graphics context font according to the task specifications.
	 */
	protected void initFont( Graphics2D gc ) {

		Font aFont;

		if ( this.useDefaultFont ) {
			if ( this.fontSize != DEFAULT_FONT_SIZE ) {
				aFont = gc.getFont();
				gc.setFont( aFont.deriveFont( (float) this.fontSize ) );
			}
			return;
		}

		// may be more efficient ways to derive the font, but this should work [DJL]
		String name = ( this.fontName != null ? this.fontName : gc.getFont().getName() );
		int sizePts = ( this.fontSize != DEFAULT_FONT_SIZE ? this.fontSize : gc.getFont().getSize() );
		aFont = new Font( name, this.fontStyle, sizePts );
		gc.setFont( aFont );

	} //initFont

	/**
	 * Load the image to be annotated.
	 */
	protected BufferedImage loadImage() 
		throws IOException
	{

		File aFile = ( this.imageSource != null ? this.imageSource : this.imageFile );

		if ( this.imageType == IMAGE_TYPE_JPEG ) {
			return this.loadJPEGImage( aFile );
		} else {
			throw new IllegalStateException( "Unknown image type: " + this.imageType );
		}

	} //loadImage

	/**
	 * Load the image contained in the specified file in JPEG format.
	 */
	protected BufferedImage loadJPEGImage( File aFile ) 
		throws IOException  /*, FileNotFoundException */
	{

		FileInputStream fs = new FileInputStream( aFile );
		JPEGImageDecoder decoder = JPEGCodec.createJPEGDecoder( fs );

		BufferedImage anImage = decoder.decodeAsBufferedImage();
		fs.close();
		return anImage;

	} //loadJPEGImage

	/**
	 * Verify that the image type is a supported format.
	 * Normalize to the constant name so we can use identity testing 
	 * for load/store dispatching during task execution.
	 */
	protected String normalizeImageType( String anImageType ) {
		String normalizedName = anImageType.toLowerCase();
		for (int i = 0; i < SUPPORTED_IMAGE_TYPES.length; i++) {
			String imageTypeConstant = SUPPORTED_IMAGE_TYPES[i]; 
			if ( anImageType.equals( imageTypeConstant ) ) {
				return imageTypeConstant;
			}
		}
		throw new BuildException( 
			"Not a supported image type: " + anImageType, 
			this.getLocation() );
	}

	/**
	 * Save the annotated image.
	 */
	protected void saveImage( BufferedImage anImage ) 
		throws IOException
	{
		if ( this.imageType == IMAGE_TYPE_JPEG ) {
			this.saveJPEGImage( anImage, this.imageFile );
		} else {
			throw new IllegalStateException( "Unknown image type: " + this.imageType );
		}
	}

	/**
	 * Save the image to the specified file in JPEG format.
	 */
	protected void saveJPEGImage( BufferedImage anImage, File aFile ) 
		throws IOException
	{

		FileOutputStream fs = new FileOutputStream( aFile ); 
		JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder( fs ); 

		encoder.encode( anImage ); 
		fs.flush(); 
		fs.close();

	} //saveJPEGImage

	/**
	 * Verify that the necessary task parameters have been specified.
	 * Fills in defaults as appropriate.
	 */
	protected void validateTaskSpecification() 
		throws BuildException
	{

		// image type is validated in the setter

        if (this.imageFile == null) {
            throw new BuildException("You must specify the image file to create!");
        }


		if ( this.fontSize != DEFAULT_FONT_SIZE ) {
			if ( this.fontSize <= 0 ) {
				throw new BuildException( "Invalid fontsize" );
			}
			// other checks on reasonableness of the font specification?!
		}

		if ( this.annotations.size() == 0 ) {
			throw new BuildException( "No image annotations specified for " + this.imageFile );
		}

	} //validateTaskSpecification


} //AnnotateImage

