/*
 * ImageAnnotation.java
 *
 * Release: 1.0
 * Build:   20021023
 *
 * Copyright 2002 Gary Gregory and Deborah Lewis. All Rights Reserved.
 *
 * This software is the proprietary information of Glaivestone Guild,
 * founded 2002 by Gary Gregory and Deborah Lewis.
 * Use is subject to license terms.
 */

package com.glaivestone.ant.taskdefs;

// Java base class library
import java.util.ArrayList;
import java.util.List;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;

// Ant
import org.apache.tools.ant.BuildException;

/**
 * Nested element in an <annotateImage> task containing
 * the specification for an annotation.
 * 
 * <p>An annotation is specified by the text line(s)
 * to display and the upper left corner within the image 
 * of the area in which the text annotation is to be displayed.
 * 
 * @author Deb Lewis
 */
public class ImageAnnotation {

    //----------------------------------------------------------------------------
    // INSTANCE VARIABLES

	protected int xLeft;
	protected int yTop;

	protected List lines;

    //----------------------------------------------------------------------------
    // CONSTRUCTORS

	public ImageAnnotation() {
		super();
		this.xLeft = -1;
		this.yTop = -1;
		this.lines = new ArrayList();
	}

    //----------------------------------------------------------------------------
    // ELEMENT ATTRIBUTE SETTERS

	public void setText( String aLine ) {
		if ( this.lines.size() > 0 ) {
			throw new BuildException(
				"text attribute not allowed when element content provided" );
		}
		this.lines.add( aLine.trim() );
	}

	/**
	 * Specify the x coordinate of the upper left corner of the annotation text area.
	 */
	public void setX( int x ) {
		this.xLeft = x;
	}

	/**
	 * Specify the y coordinate of the upper left corner of the annotation text area.
	 */
	public void setY( int y ) {
		this.yTop = y;
	}

    //----------------------------------------------------------------------------
    // ELEMENT CONTENT SETTERS

	/**
	 * Parse text content into one or more lines of annotation text.
	 */
	public void addText( String text ) {

		BufferedReader textReader = new BufferedReader( new StringReader( text ) );
		String aLine;

		try {
			while ( ( aLine = textReader.readLine() ) != null ) {
				aLine = aLine.trim();
				if ( aLine.length() > 0) {
					this.lines.add( aLine );
				}
			}
		} catch( IOException ex ) {
			throw new BuildException( "I/O error parsing annotation text", ex );
		}

	} //addText

    //----------------------------------------------------------------------------
    // ACCESSORS

	public int getLeft() {
		return this.xLeft;
	}

	public int getTop() {
		return this.yTop;
	}

	public List getLines() {
		return this.lines;
	}

	public boolean hasText() {
		int nLines = this.lines.size();
		int nValidLines = 0;
		for (int i = 0; i < nLines; i++) {
			String aLine = (String) this.lines.get( i );
			if ( aLine.length() > 0 ) {
				nValidLines++;
			}
		}
		return ( nValidLines > 0 );
	}

} //ImageAnnotation

